/*
linux/fs/dmsdos/dmsdos_read.c

DMSDOS filesystem: read access routines.

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/msdos_fs.h>
#include <linux/dmsdos_fs.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <asm/system.h>

#define PRINTK(X)
#define CLPOS 26

extern int dbl_cvf_inos[];
extern Dblsb dblsb[];

/* returns ino or -ENOENT */
/* must not break on empty/zero entries */
int read_dbl_direntry(struct super_block*sb,
                      int dbl_clusterno,int cvfnr,int entrynr,
                      unsigned char*buf)
{ int dbl_sector;
  int dbl_offset;
  int ino;
  struct buffer_head*bh;
  int i;
  Mdfat_entry mde;
  
  /* Hmm ...  error test: */
  if(entrynr>65535*16*dblsb[cvfnr].s_sectperclust)
  { /* impossible, there are max 65535 clusters a 16*sectperclust entries */
    /* this is most likely a fat loop (code prevents hang) */
    printk("DMSDOS: read_dbl_direntry: entrynr too large, aborting!\n");
    return -ENOENT;
  }
  
  /* get sector and offset */
  if(dbl_clusterno==0) /*it is a dbl rootdir*/
  { dbl_sector=dblsb[cvfnr].s_rootdir+(entrynr>>4);
    if(entrynr>=dblsb[cvfnr].s_rootdiranzentry)return -ENOENT;
  }
  else
  { while(entrynr/(16*dblsb[cvfnr].s_sectperclust))
    { entrynr-=(16*dblsb[cvfnr].s_sectperclust);
      dbl_clusterno=dbl_fat_nextcluster(sb,dbl_clusterno,cvfnr,NULL);
      if(dbl_clusterno==-1)return -ENOENT;
      if(entrynr<0)
      { printk("DMSDOS: read_dbl_direntry: error calculating clusterno\n");
        return -ENOENT;
      }
    }
    
    dbl_mdfat_value(sb,dbl_clusterno,cvfnr,NULL,&mde);
    
    /* handle compressed directory clusters under drivespace 3 */  
    if((mde.flags&1)==0&&64==dblsb[cvfnr].s_sectperclust)
    { printk("DMSDOS: read_dbl_direntry: compressed drivespace 3 directory cluster, can't read!"
             " clusternr=%d flags=%d sizehi=%d cvfnr=%d\n",
             dbl_clusterno,mde.flags,mde.size_hi_minus_1+1,cvfnr+1);
      return -EIO;
    }
    
    dbl_sector=mde.sector_minus_1+1; /*dbl_mdfat_cluster2sector(sb,dbl_clusterno,cvfnr);*/
    /* end of shortened dir cluster reached ? */
    if((entrynr>>4)>mde.size_hi_minus_1)return -ENOENT;
    dbl_sector+=entrynr>>4;
  }
  /* always the same */
  dbl_offset=entrynr&0xf;
  ino=((cvfnr+1)<<DMSDOS_CVFNR_BITS)+(dbl_sector<<4)+dbl_offset;
  /* read sector */
  bh=read_dbl_sector(sb,dbl_sector,cvfnr);
  if(bh==NULL)return -EIO;
  for(i=0;i<32;++i)buf[i]=bh->b_data[i+32*dbl_offset];
  bh_free(sb,bh);
  return ino;
}

int scan_dbl_dir_4_clusterno(struct super_block*sb,
                             int dirstartclust, int tofind,int cvfnr)
{ int ino;
  int ent=0;
  unsigned char buf[32];
  unsigned char * pp;

  while((ino=read_dbl_direntry(sb,dirstartclust,cvfnr,ent,buf))>0)
  { ++ent;
    pp=&(buf[CLPOS]);
    /* break on zero entry */
    if(buf[0]==0)return -ENOENT;
    if(buf[0]!=0xe5&&(CHS(pp)==tofind))return ino;
  }
  return ino;
}

/* does *not* look for . or ..
   len==-1 means name is already raw msdos formatted */
int scan_dbl_dir_4_filename(struct super_block*sb, int dirstartclust,
                            unsigned const char*tofind,int len,
                            int cvfnr)
{ int ino;
  int ent=0;
  unsigned char buf[32];
  unsigned char msdosname[]="           ";
  int i;
  int j;
  int warpunkt=0;
  
  if(len<0)strncpy(msdosname,tofind,11);
  else
  {
  
  for(i=0,j=0;i<len&&j<11;++i)
  { if(tofind[i]=='.')
    { if(j>8)return -ENOENT;
      if(warpunkt)return -ENOENT;
      warpunkt=1;
      j=8;
      continue;
    }
    if(tofind[i]>='a'&&tofind[i]<='z')msdosname[j++]=tofind[i]-'a'+'A';
    else msdosname[j++]=tofind[i];
  }
  
  }
/*  
  printk("DMSDOS: scan_dbl_dir_4_filename: name=%s dosname=%s\n",
         tofind,msdosname);
  printk("DMSDOS: scan_dbl_dir_4_filename: dirstartclust=%d cvfnr=%d\n",
         dirstartclust,cvfnr+1);
*/
  while((ino=read_dbl_direntry(sb,dirstartclust,cvfnr,ent,buf))>0)
  { ++ent;
    /* break on zero entry */
    if(buf[0]==0)return -ENOENT;
    if(strncmp(buf,msdosname,11)==0)
    { /*printk("DMSDOS: scan_dbl_dir_4_filename: found with ino=%d\n",ino);*/
      return ino;
    }
  }
/*
  printk("DMSDOS: scan_dbl_dir_4_filename: not found, error=%d, ent=%d\n",
         ino,ent);
*/
  return ino;
}

/* Returns the inode number of the directory entry at offset pos.
   Pos is incremented. */

int dmsdos_get_entry(struct inode *dir, loff_t *pos,unsigned char * buf)
{
	int entrynr;
	int ino;
	int clust;
	int cvfnr;

		entrynr = (*pos)>>5;
		*pos += 32;
		if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
		else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
		clust=MSDOS_I(dir)->i_start;/*dbl_inode2startcluster(dir);*/
		ino=read_dbl_direntry(dir->i_sb,clust,cvfnr,entrynr,buf);
		/* break on zero entry */
		if(buf[0]==0)return -ENOENT;
		return ino;
}

int dmsdos_parent_ino(struct inode*dir)
{  struct buffer_head*bh;
   int cvfnr;
   int parent_dbl_cluster;
   int parent_dbl_sector;
   int parent_parent_dbl_cluster;
   unsigned char * pp;

   if(dbltest(dir))return MSDOS_ROOT_INO;

   cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
   if(cvfnr<0)return -ENOENT;

/* read first sector of my own cluster */
   bh=read_dbl_sector(dir->i_sb,
                      dbl_mdfat_cluster2sector(dir->i_sb,
                                               MSDOS_I(dir)->i_start,cvfnr),
                      cvfnr);
   if(bh==NULL)return -EIO;
/* and get second entry (it is ..) */
   pp=&(bh->b_data[32+CLPOS]);
   parent_dbl_cluster=CHS(pp);
   bh_free(dir->i_sb,bh);
    
   if(parent_dbl_cluster==0)
   { /* Verweis auf dbl_rootdir */
     return dbl_cvf_inos[cvfnr];
   }
   
   parent_dbl_sector=dbl_mdfat_cluster2sector(dir->i_sb,parent_dbl_cluster,
                                              cvfnr);
   
   bh=read_dbl_sector(dir->i_sb,parent_dbl_sector,cvfnr);
   if(bh==NULL)return -EIO;
   pp=&(bh->b_data[32+CLPOS]);
   parent_parent_dbl_cluster=CHS(pp);
   bh_free(dir->i_sb,bh);
   return scan_dbl_dir_4_clusterno(dir->i_sb,
                         parent_parent_dbl_cluster, /*dir start cluster*/
                         parent_dbl_cluster,        /* to find */ 
                         cvfnr);
}

#ifdef __FOR_KERNEL_1_3
/* kernel 1.3.xx needs a new readdir function .... I hope I did not break
   anything in dmsdos' readdir */

int dmsdos_readdirx(
	struct inode *inode,
	struct file *filp,
	void *dirent,
	filldir_t filldir)
{
	int ino,i /*,i2,last*/ ;
	/*char c;*/
	unsigned char buf[32];
	struct msdos_dir_entry *de;
	unsigned long oldpos = filp->f_pos;
	char longname[257];

	if (!inode || !S_ISDIR(inode->i_mode))
		return -EBADF;
/* Fake . and .. for the root directory. */
	if (dbltest(inode)) {
		while (oldpos < 2) {
			if (filldir(dirent, "..", oldpos+1, oldpos, MSDOS_ROOT_INO) < 0)
				return 0;
			oldpos++;
			filp->f_pos++;
		}
		if (oldpos == 2)
			filp->f_pos = 0;
	}
	if (filp->f_pos & (sizeof(struct msdos_dir_entry)-1))
		return -ENOENT;
	while ((ino = dmsdos_get_lfn_entry(inode,&filp->f_pos,buf,
	                                   longname)) > -1) {
	        de=(struct msdos_dir_entry*) buf;
		if (!IS_FREE(de->name)
		        /* this is copied from the new msdos readdir -
		           ignore vol, ok
		           ignore sys, hmm...
		           but why ignore hidden ???
		           well, doesn't matter, compressed ones don't
		           run through this code here....
		           but I do not like ignoring hidden files!
		        */
			&& !(de->attr & (ATTR_VOLUME))) {
		/*	
			char bufname[12];
			char *ptname = bufname;
			for (i = last = 0; i < 8; i++) {
				if (!(c = de->name[i])) break;
				if (c >= 'A' && c <= 'Z') c += 32;
				if (c != ' ')
					last = i+1;
				ptname[i] = c;
			}
			i = last;
			ptname[i] = '.';
			i++;
			for (i2 = 0; i2 < 3; i2++) {
				if (!(c = de->ext[i2])) break;
				if (c >= 'A' && c <= 'Z') c += 32;
				if (c != ' ')
					last = i+1;
				ptname[i] = c;
				i++;
			}
			if ((i = last) != 0) {
		*/
		        i=strlen(longname);
		        if(i!=0) {
				if (!strcmp(de->name,MSDOS_DOT))
					ino = inode->i_ino;
				else if (!strcmp(de->name,MSDOS_DOTDOT))
				      {
					ino = dmsdos_parent_ino(inode);
					if(ino<0)return ino;
				      }
				if (filldir(dirent, longname, i, oldpos, ino) < 0) {
					filp->f_pos = oldpos;
					break;
				}
			}
		}
		oldpos = filp->f_pos;
	}
	return 0;
}

#else
/* kernel 1.2.xx readdir function */

#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
#define ROUND_UP(x) (((x)+3) & ~3)

int dmsdos_readdirx( 
	struct inode *inode,
	struct file *filp,
	struct dirent *dirent,	/* dirent in user space */
	int count)
{	
	int ino,i /*,i2,last*/ ;
	char /*c,*/ *walk;
	unsigned char buf[32];
	struct msdos_dir_entry *de;
	char longname[257];

	if (!inode || !S_ISDIR(inode->i_mode)) return -EBADF;
	if (dbltest(inode)) {
	        /* printk("DMSDOS: readdirx reading dblrootdir ino=%ld\n",
	                                                     inode->i_ino);
	        */
/* Fake . and .. for the root directory. */
		if (filp->f_pos == 2) filp->f_pos = 0;
		else if (filp->f_pos < 2) {
			walk = filp->f_pos++ ? ".." : ".";
			for (i = 0; *walk; walk++)
				put_fs_byte(*walk,dirent->d_name+i++);
			put_fs_long(filp->f_pos==1?inode->i_ino:MSDOS_ROOT_INO,&dirent->d_ino);
			put_fs_byte(0,dirent->d_name+i);
			put_fs_word(i,&dirent->d_reclen);
			return ROUND_UP(NAME_OFFSET(dirent) + i + 1);
		}
	}
	/*else printk("DMSDOS: readdirx ??? ino=%ld\n",inode->i_ino);*/
	if (filp->f_pos & (sizeof(struct msdos_dir_entry)-1)) return -ENOENT;

	while ((ino = dmsdos_get_lfn_entry(inode,&filp->f_pos,buf,
	                                   longname)) > -1) {
	        de=(struct msdos_dir_entry*) buf;
		if (!IS_FREE(de->name) && !(de->attr & ATTR_VOLUME)) {
	/*
			char bufname[13];
			char *ptname = bufname;
			for (i = last = 0; i < 8; i++) {
				if (!(c = de->name[i])) break;
				if (c >= 'A' && c <= 'Z') c += 32;
				if (c != ' ')
					last = i+1;
				ptname[i] = c;
			}
			i = last;
			ptname[i] = '.';
			i++;
			for (i2 = 0; i2 < 3; i2++) {
				if (!(c = de->ext[i2])) break;
				if (c >= 'A' && c <= 'Z') c += 32;
				if (c != ' ')
					last = i+1;
				ptname[i] = c;
                               i++;
			}
	*/
			if ((i = strlen(longname)) != 0) {	
				if (!strcmp(de->name,MSDOS_DOT))
					ino = inode->i_ino;
				else if (!strcmp(de->name,MSDOS_DOTDOT))
				     {
				       ino = dmsdos_parent_ino(inode);
			               if(ino<0)return ino;
			             }
				/*bufname[i] = '\0'*/;
				put_fs_long(ino,&dirent->d_ino);
				memcpy_tofs(dirent->d_name,longname,i+1);
				put_fs_word(i,&dirent->d_reclen);
				
				return ROUND_UP(NAME_OFFSET(dirent) + i + 1);
			}
		}
	}
	return 0;
}

#endif

int dmsdos_file_readx(
	struct inode *inode,
	struct file *filp,
	char *buf,
	int count)
{      
        int clusternr;
        unsigned char*clusterd;
        int cvfnr;
        int offset;
        int gelesen;
        int membytes;
        int ret;
        char * b;
        int toread;
        char datum;
        int need_cluster;

	if (!inode) {
		printk("dmsdos_file_read: inode = NULL\n");
		return -EINVAL;
	}
	/* S_ISLNK allows for UMSDOS. Should never happen for normal MSDOS */
	if (!S_ISREG(inode->i_mode) && !S_ISLNK(inode->i_mode)) {
		printk("dmsdos_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}

        cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
        if(cvfnr<0)
        { printk("DMSDOS: file_readx: invalid cvfnr ino=%ld\n",inode->i_ino);
          return -EIO;
        }
        
        if(count<=0)return 0;
        
        if(filp->f_pos>=inode->i_size)return 0;
        
        if(filp->f_pos+count>inode->i_size)count=inode->i_size-filp->f_pos;
        
        membytes=SECTOR_SIZE*dblsb[cvfnr].s_sectperclust;
        
        /* calculate clusternr for cluster to read */
        clusternr=MSDOS_I(inode)->i_start;
        offset=filp->f_pos;
        while(offset>=membytes&&clusternr>0)
        {  offset-=membytes;
           clusternr=dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,NULL);
        }
        if(clusternr<=0)
        {  printk("DMSDOS: file_readx: FAT mismatches file size for ino=%ld\n",
                   inode->i_ino);
           return 0;        
        }
        
        gelesen=0;
        b=buf;
        
        clusterd=(unsigned char*)MALLOC(membytes);
        if(clusterd==NULL)
        { printk("DMSDOS: file_readx: no memory!\n");
          return -EIO;
        }
        
        if(MSDOS_I(inode)->i_binary==0)goto text_read;
        
        do
        {  ret=dmsdos_read_cluster(inode->i_sb,clusterd,clusternr,cvfnr);
           if(ret==0)
           { toread=(membytes-offset>count) ? count : membytes-offset;
             /*printk("DMSDOS file_readx: memcpy_tofs(0x%08x,0x%08x,0x%08x)\n",
                     b,clusterd+offset,toread);*/
             memcpy_tofs(b,clusterd+offset,toread);
             gelesen+=toread;
             count-=toread;
             if(count>0)
             { b+=toread;
               offset=0;
               clusternr=dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,NULL);
               if(clusternr<=0)
               { ret=-1;
                 printk("DMSDOS: file_readx: FAT mismatches file size for ino=%ld\n",
                         inode->i_ino);
               }
             }
           }
        }
        while(count>0&&ret==0);
        
        FREE(clusterd);
        
        filp->f_pos+=gelesen;
        return gelesen;
        
text_read:
        /* ok, let's do it byte for byte..... */
        gelesen=0;
        need_cluster=1;
        while(count>0&&inode->i_size>filp->f_pos)
        { if(need_cluster)
          { ret=dmsdos_read_cluster(inode->i_sb,clusterd,clusternr,cvfnr);
            if(ret<0)
            { FREE(clusterd);
              return -EIO;
            }
            clusternr=dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,NULL);
            need_cluster=0;
          }

          datum=clusterd[offset++];

          ++(filp->f_pos);
          if(datum!=13)
          { put_fs_byte(datum,&(buf[gelesen]));
            ++gelesen;
            --count;
          }
          if(offset==membytes)
          { need_cluster=1;
            offset=0;
          }
        }
        
        FREE(clusterd);
        return gelesen;
}

int dmsdos_subdirs(struct inode*inode)
{ int cvfnr;
  int startclust;
  int ent=0;
  int res=0;
  unsigned char buf[32];
  
  /*printk("DMSDOS: subdirs, ino=%ld\n",inode->i_ino);*/
  
  if((cvfnr=dbltest(inode))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
  else cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
  startclust=MSDOS_I(inode)->i_start;
  
  while(read_dbl_direntry(inode->i_sb,startclust,cvfnr,ent,buf)>0)
  { ++ent;
    if(buf[0]!=0xe5&&buf[0]!=0&&(buf[11]&0x10)!=0)++res;
  }
  return res;
}

void dmsdos_read_inodex(struct inode *inode)
{	
        int dblsec;
        int offs;
        int cvfnr;
        struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *raw_entry;
	int nr;
	int zaehler;

/* printk("read inode %d\n",inode->i_ino); */
	MSDOS_I(inode)->i_busy = 0;
	MSDOS_I(inode)->i_depend = MSDOS_I(inode)->i_old = NULL;
	MSDOS_I(inode)->i_binary = 1;
#ifdef __FOR_KERNEL_1_3_8x
        inode->i_uid = MSDOS_SB(inode->i_sb)->options.fs_uid;
        inode->i_gid = MSDOS_SB(inode->i_sb)->options.fs_gid;
#else
	inode->i_uid = MSDOS_SB(inode->i_sb)->fs_uid;
	inode->i_gid = MSDOS_SB(inode->i_sb)->fs_gid;
#endif
	if ((cvfnr=dbltest(inode))!=0) {
	        cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
		inode->i_mode = ((S_IRUGO | S_IXUGO | 
		       ( /*dblsb[cvfnr].s_comp==READ_ONLY ? 0 :*/ S_IWUGO ) )& 
		       ~MSDOS_SB(inode->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
                             options.fs_umask
#else
		             fs_umask
#endif
		                ) | S_IFDIR;
		inode->i_op = &dmsdos_dir_inode_operations;
		MSDOS_I(inode)->i_start=0;/*muss bekannt sein fr _subirs*/
		inode->i_nlink = dmsdos_subdirs(inode)+2;
		    /* subdirs (neither . nor ..) plus . and "self" */
		inode->i_size = dblsb[cvfnr].s_rootdiranzentry*
		    sizeof(struct msdos_dir_entry);
		inode->i_blksize = dblsb[cvfnr].s_sectperclust*
		    SECTOR_SIZE;
		inode->i_blocks = (inode->i_size+inode->i_blksize-1)/
		    inode->i_blksize*dblsb[cvfnr].s_sectperclust;
		MSDOS_I(inode)->i_attrs = 0;
		
		/* read it */
		if(!(bh=read_real_sector(inode->i_sb,inode->i_ino>>4)))
		{
			printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
			panic("dmsdos_read_inode: unable to read i-node block");
		}
		raw_entry = &((struct msdos_dir_entry *) (bh->b_data))
	    		              [inode->i_ino & (MSDOS_DPB-1)];
		inode->i_mtime = inode->i_atime = inode->i_ctime = 
		        date_dos2unix(CF_LE_W(raw_entry->time),
		                      CF_LE_W(raw_entry->date));
		return;
	}
	
	dblsec=(inode->i_ino&DMSDOS_ORIGINO_MASK)>>4;
	offs=inode->i_ino&0xf;
	cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;

	if (!(bh = read_dbl_sector(inode->i_sb,dblsec,cvfnr))) {
		printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
		panic("dmsdos_read_inode: unable to read i-node block");
	}
	raw_entry = &((struct msdos_dir_entry *) (bh->b_data))
	    [inode->i_ino & (MSDOS_DPB-1)];
	if ((raw_entry->attr & ATTR_DIR) && !IS_FREE(raw_entry->name)) {
#ifdef __FOR_KERNEL_1_3_8x
		inode->i_mode = MSDOS_MKMODE(raw_entry->attr,(S_IRUGO|S_IXUGO|
		              (/*dblsb[cvfnr].s_comp==READ_ONLY ? 0 :*/ S_IWUGO) )
		              & ~MSDOS_SB(inode->i_sb)->
		                    options.fs_umask
		                         ) | S_IFDIR;
#else
		inode->i_mode = MSDOS_MKMODE(raw_entry->attr,(S_IRUGO|S_IXUGO|
		              (/*dblsb[cvfnr].s_comp==READ_ONLY ? 0 :*/ S_IWUGO) )
		              & ~MSDOS_SB(inode->i_sb)->
		                    fs_umask
		                         ) | S_IFDIR;
#endif
		inode->i_op = &dmsdos_dir_inode_operations;
		MSDOS_I(inode)->i_start = CF_LE_W(raw_entry->start);
		inode->i_nlink = dmsdos_subdirs(inode);
		    /* includes .., compensating for "self" */
#ifdef DEBUG
		if (!inode->i_nlink) {
			printk("directory %d: i_nlink == 0\n",inode->i_ino);
			inode->i_nlink = 1;
		}
#endif
		inode->i_size = 0;
		
		zaehler=0;
		if ((nr = CF_LE_W(raw_entry->start)) != 0)
			while (nr != -1) {
				inode->i_size += 
				            dblsb[cvfnr].s_sectperclust*512;
				if (!(nr = dbl_fat_nextcluster(inode->i_sb,
				                               nr,cvfnr,
				                               NULL))) {
					printk("DMSDOS: readinodex: bad FAT (unexpected zero), ino=%ld\n",inode->i_ino);
					break;
				}
				
				++zaehler;
				if(zaehler>65535)
				{ printk("DMSDOS: read_inodex: loop in FAT, ino=%ld\n",inode->i_ino);
				  break;
				}
			}
	}
	else {
#ifdef __FOR_KERNEL_1_3_8x
		inode->i_mode = MSDOS_MKMODE(raw_entry->attr,((IS_NOEXEC(inode)
		    ? S_IRUGO : S_IRUGO|S_IXUGO) |
		        (/*dblsb[cvfnr].s_comp==READ_ONLY ? 0 :*/ S_IWUGO) ) &
		        ~MSDOS_SB(inode->i_sb)->
                               options.fs_umask
		                 ) | S_IFREG;
#else
		inode->i_mode = MSDOS_MKMODE(raw_entry->attr,((IS_NOEXEC(inode)
		    ? S_IRUGO : S_IRUGO|S_IXUGO) |
		        (/*dblsb[cvfnr].s_comp==READ_ONLY ? 0 :*/ S_IWUGO) ) &
		        ~MSDOS_SB(inode->i_sb)->
		               fs_umask
		                 ) | S_IFREG;
#endif
		inode->i_op = sb->s_blocksize == 1024
			? &dmsdos_file_inode_operations_1024
			: &dmsdos_file_inode_operations;
		MSDOS_I(inode)->i_start = CF_LE_W(raw_entry->start);
		inode->i_nlink = 1;
		inode->i_size = CF_LE_L(raw_entry->size);
	}
	MSDOS_I(inode)->i_binary = is_binary(MSDOS_SB(inode->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
                   options.conversion,
#else          	
	           conversion,
#endif	    
	    raw_entry->ext);
	MSDOS_I(inode)->i_attrs = raw_entry->attr & ATTR_UNUSED;
	/* this is as close to the truth as we can get ... */
	inode->i_blksize = dblsb[cvfnr].s_sectperclust*SECTOR_SIZE;
	inode->i_blocks = (inode->i_size+inode->i_blksize-1)/
	    inode->i_blksize*dblsb[cvfnr].s_sectperclust;
	inode->i_mtime = inode->i_atime = inode->i_ctime =
	    date_dos2unix(CF_LE_W(raw_entry->time),CF_LE_W(raw_entry->date));
	bh_free(inode->i_sb,bh);

        /*printk("DMSDOS: readinodex reading inode ino=%ld\n",inode->i_ino);*/
}

int dmsdos_lookupx(struct inode *dir,const char *name,int len,
    struct inode **result)
{
	int ino;
	int cvfnr;
	int startclust;
	struct inode*next;
	
	PRINTK (("dmsdos_lookup\n"));

	*result = NULL;
	if (!dir) return -ENOENT;
	if (!S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}
	PRINTK (("dmsdos_lookup 2\n"));
	if (len == 1 && name[0] == '.') {
		*result = dir;
		return 0;
	}
	if (len == 2 && name[0] == '.' && name[1] == '.') {
	        if(dbltest(dir))
	        { /*printk("DMSDOS: lookupx '..' result MDSOS_ROOT\n");*/
	          ino=MSDOS_ROOT_INO;
	        }
		else
		{ /*printk("DMSDOS: lookupx '..' inode ino=%ld\n",dir->i_ino);*/
		  ino=dmsdos_parent_ino(dir);
		}
		iput(dir);
		if (ino < 0) return ino;
		if (!(*result = iget(dir->i_sb,ino))) return -EACCES;
		return 0;
	}
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	startclust=MSDOS_I(dir)->i_start;/*dbl_inode2startcluster(dir);*/
	ino=scan_dbl_dir_4_lfn(dir->i_sb,startclust,name,len,cvfnr,NULL);
	if(ino<0)
	{ iput(dir);
	  return ino;
	}
	if(!(*result=iget(dir->i_sb,ino))) return -EACCES;
	
	PRINTK (("dmsdos_lookup 5\n"));
	if (MSDOS_I(*result)->i_busy) { /* mkdir in progress */
		iput(*result);
		iput(dir);
		return -ENOENT;
	}
	PRINTK (("dmsdos_lookup 6\n"));
	while (MSDOS_I(*result)->i_old) {
		next = MSDOS_I(*result)->i_old;
		iput(*result);
		if (!(*result = iget(next->i_sb,next->i_ino))) {
#ifdef __FOR_KERNEL_1_3_8x
			fat_fs_panic(dir->i_sb,"msdos_lookup: Can't happen");
#else
			fs_panic(dir->i_sb,"msdos_lookup: Can't happen");
#endif
			iput(dir);
			return -ENOENT;
		}
	}
	PRINTK (("dmsdos_lookup 7\n"));
	iput(dir);
	PRINTK (("dmsdos_lookup 8\n"));
	return 0;
}
