/*
linux/fs/dmsdos/dmsdos_write.c

DMSDOS filesystem: write access functions

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include <asm/segment.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/dmsdos_fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/locks.h>

#define PRINTK(X)

#include <asm/segment.h>
#include <linux/fcntl.h>

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

#define RAW -1

extern Dblsb dblsb[];

/*
	Write to a file from user space
*/
int dmsdos_file_writex(struct inode *inode, struct file *filp, 
#ifdef __FOR_KERNEL_1_3
                       const
#endif
                       char *buf, int count)
{       int cluster;
        int lastcluster;
        int sektor;
        Mdfat_entry mde;
        int new;
        unsigned int offset;
        int cvfnr;
        unsigned char *clusterd;
#ifdef __FOR_KERNEL_1_3
        const
#endif
        unsigned char *b;
        int length;
        int size;
        int canwrite;
        int geschrieben;
        int clustersize;
        int uc;
        int cr;
        char datum;
        int need_cluster;

	if (!inode) {
		printk("dmsdos_file_write: inode = NULL\n");
		return -EINVAL;
	}
	/* S_ISLNK allows for UMSDOS. Should never happen for normal MSDOS */
	if (!S_ISREG(inode->i_mode) && !S_ISLNK(inode->i_mode)) {
		printk("dmsdos_file_write: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
/*
 * ok, append may not work when many processes are writing at the same time
 * but so what. That way leads to madness anyway.
 */
        cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
        if(cvfnr<0)
        { printk("DMSDOS: file_writex: cvfnr<0 ???\n");
          return -EINVAL;
        }
        
        if(dblsb[cvfnr].s_comp==READ_ONLY)return -EPERM;
 
	if (filp->f_flags & O_APPEND) filp->f_pos = inode->i_size;
	if (count <= 0) return 0;
	
	clustersize=dblsb[cvfnr].s_sectperclust*SECTOR_SIZE;
	uc=(MSDOS_I(inode)->i_binary>1)?1:0; /* uncompressed forced ? */
	
	clusterd=(unsigned char*)MALLOC(clustersize);
	if(clusterd==NULL)
	{ printk("DMSDOS: write_file: no memory!\n");
	  return -EIO;
	}
	
	offset=filp->f_pos;
	/* bestimme cluster */
	new=0;
	lastcluster=0;
	cluster=MSDOS_I(inode)->i_start;
	if(cluster==0)
	{ /*ersten cluster holen*/
	  cluster=allocate_first_cluster(inode->i_sb,cvfnr);
	  if(cluster<=0)
	  { FREE(clusterd);
	    return -ENOSPC;
	  }
	  MSDOS_I(inode)->i_start=cluster;
	  inode->i_dirt=1;
	  new=1;
	}
	while(offset>=clustersize)
	{ lastcluster=cluster;
	  cluster=dbl_fat_nextcluster(inode->i_sb,cluster,cvfnr,NULL);
	  if(cluster==0||cluster==1)
	  { printk("DMSDOS: file_write: calc_pos: something wrong with fat (0 or 1) in cluster %d\n",
	           lastcluster);
	    FREE(clusterd);
	    return -EIO;
	  }
	  if(cluster==-1)
	  { /* war der letzte, ich brauche einen neuen */
	    /*printk("DMSDOS: write_file: calc_pos: need new cluster after cluster %d\n",
	           lastcluster);*/
	    /* hat der letzte die volle laenge ? */
	    dbl_mdfat_value(inode->i_sb,lastcluster,cvfnr,NULL,&mde);
	    size=mde.size_hi_minus_1+1;
	    if(size!=dblsb[cvfnr].s_sectperclust)
	    { /*nein, mist....*/
	      /*printk("DMSDOS: write_file: calc_pos: adapting cluster size cluster=%d size=%d\n",
	             lastcluster,size);*/
	      if(dmsdos_read_cluster(inode->i_sb,clusterd,lastcluster,cvfnr)<0)
	      { FREE(clusterd);
	        return -EIO;
	      }
	      memset(&(clusterd[SECTOR_SIZE*size]),0,
	             (dblsb[cvfnr].s_sectperclust-size)*SECTOR_SIZE);
	      /*schreiben(near mdfat&DBLSECTOR_MASK+1)*/
	      if( dmsdos_write_cluster(inode->i_sb,clusterd,clustersize,
	                               lastcluster,mde.sector_minus_1+1,
	                               cvfnr,
	                               uc) <0 )
	      { FREE(clusterd);
	        return -ENOSPC;
	      }
	    }
	    sektor=dbl_mdfat_cluster2sector(inode->i_sb,lastcluster,cvfnr);
	    /*printk("DMSDOS: write_file: calc_pos: allocating new cluster...\n");
	    */
	    cluster=allocate_next_cluster(inode->i_sb,lastcluster,cvfnr);
	    /*printk("DMSDOS: write_file: calc_pos: new cluster=%d\n",cluster);
	    */
	    if(cluster<=0)
	    { FREE(clusterd);
	      return -ENOSPC;
	    }
	    /* wuerde noch ein durchlauf folgen, muss die luecke gefuellt 
	       werden - ansonsten ists egal */
	    if(offset-clustersize>=clustersize)
	    {
	      /*ganz mit nullen fuellen*/
	      /*printk("DMSDOS: zeroing new cluster...\n");*/
	      memset(clusterd,0,clustersize);
	      /*schreiben(near sektor)*/
	      /*printk("DMSDOS: writing zeroed cluster %d...\n",cluster);*/
	      dmsdos_write_cluster(inode->i_sb,clusterd,clustersize,cluster,
	                           sektor,cvfnr,uc);
	      /*printk("DMSDOS: cluster written\n");*/
	    }
	    new=1;
	  }
	  offset-=clustersize;
	}
	 
	/*printk("DMSDOS: file_write: beginning with cluster %d, lastcluster=%d, new=%d\n",
	       cluster,lastcluster,new);*/
	  /* in cluster steht die nummer des clusters mit dem ich jetzt
	     arbeite;
	     der cluster existiert wenn new=0 und wurde evtl mit
	     nullen ganz aufgefuellt;
	     es ist egal, ob es der letzte in der fatkette ist;
	  */
        b=buf;
        geschrieben=0;
        
        if(MSDOS_I(inode)->i_binary==0)goto text_write;
        
        while(count>0)
        {
	  if( (offset>0||count<clustersize) &&new==0)
	  { /* muss gelesen werden */
	    /*printk("DMSDOS: write_file: reading cluster %d...\n",cluster);*/
	    dbl_mdfat_value(inode->i_sb,cluster,cvfnr,NULL,&mde);
	    length=(mde.size_hi_minus_1+1)*SECTOR_SIZE;
	    if(dmsdos_read_cluster(inode->i_sb,clusterd,cluster,cvfnr)<0)
	    { printk("DMSDOS: write_file: read_cluster failed!\n");
	      FREE(clusterd);
	      return -EIO;
	    }
          }
          else length= (new==1&&count+offset<clustersize) ?
                                                  count+offset : clustersize;
	
	  canwrite=MIN(clustersize-offset,count);
	  memcpy_fromfs(&(clusterd[offset]),b,canwrite);
	  sektor= lastcluster==0 ? 0 :
	             dbl_mdfat_cluster2sector(inode->i_sb,lastcluster,cvfnr);
	             
	  /* habe ich ueber das ende hinausgeschrieben ? */
	  if(canwrite+offset>length)
	  { /*printk("DMSDOS: write_file: write beyond file end, appending.\n");
	    */
	    length=canwrite+offset;
	  }
	  if(length>clustersize)
	  { printk("DMSDOS: write_file: length>clustersize ??? bug !!!\n");
	    length=clustersize; 
	  }
	   
	  /*printk("DMSDOS: write_file: writing cluster %d...\n",cluster);*/
	  if(dmsdos_write_cluster(inode->i_sb,clusterd,length,cluster,
	                          sektor,cvfnr,uc)<0)
	    { printk("DMSDOS: write_file: cluster write failed!\n");
	      break;
	    }
	                                                 
	  offset=0;
	  b+=canwrite;
	  geschrieben+=canwrite;
	  count-=canwrite;
	  
	  if(count==0)break; /* braucht keinen neuen cluster mehr*/
	  
	  /* naechster cluster ? */
	  lastcluster=cluster;
	  cluster=dbl_fat_nextcluster(inode->i_sb,cluster,cvfnr,NULL);
	  if(cluster==-1)
	  { 
	    /*printk("DMSDOS: write_file: write_loop: allocating new cluster after cluster %d...\n",
	           lastcluster);*/
	    cluster=allocate_next_cluster(inode->i_sb,lastcluster,cvfnr);
	    /*printk("DMSDOS: write_file: write_loop: new cluster=%d\n",
	             cluster);*/
	    if(cluster<=0) break;
	    new=1;
	  }
	}
	
	FREE(clusterd);
	filp->f_pos+=geschrieben;
	
	if(filp->f_pos>inode->i_size)
	{ inode->i_size=filp->f_pos;
	  inode->i_dirt=1;
	}
	
	return geschrieben;
		
text_write:
        geschrieben=0;
        need_cluster=1;
        cr=0;
        length=0; /* avoids warning */
        while(count>0||cr==1)
        { if(need_cluster)
          { /*printk("DMSDOS: text_write: need_cluster %d new=%d mdfat=0x%x\n",
                   cluster,new,
                   dbl_mdfat_value(inode->i_sb,cluster,cvfnr,NULL));
            */
            if(new)   /* new means cluster is new and wasn't written before*/
            { length=SECTOR_SIZE; 
              new=0;
            }
            else
            {
              dbl_mdfat_value(inode->i_sb,cluster,cvfnr,NULL,&mde);
              length=(mde.size_hi_minus_1+1)*SECTOR_SIZE;
	      if(dmsdos_read_cluster(inode->i_sb,clusterd,cluster,cvfnr)<0)
	      { printk("DMSDOS: write_file: read_cluster failed!\n");
	        FREE(clusterd);
	        return -EIO;
	      }
	    }
	    need_cluster=0;
          }
          
          if(cr==1)
          { datum=10;
            cr=0;
          }
          else
          { datum=get_fs_byte(&(buf[geschrieben]));
            ++geschrieben;
            if(datum==10)
            { datum=13;
              cr=1;
            }
            --count;
          }

          clusterd[offset++]=datum;

          ++(filp->f_pos);
          if(offset>length)length=offset;
          
          if(offset==clustersize)
          { sektor= lastcluster==0 ? 0 :
	             dbl_mdfat_cluster2sector(inode->i_sb,lastcluster,cvfnr);

            if(dmsdos_write_cluster(inode->i_sb,clusterd,length,cluster,
	                          sektor,cvfnr,uc)<0)
	    { printk("DMSDOS: write_file: cluster write failed!\n");
	      break;
	    }
	    /*check whether end reached */
	    if(count==0&&cr==0)
	    { offset=0; /* tells that there's no rest */
	      break;
	    }
            /* check whether a new cluster is needed */
            lastcluster=cluster;
            cluster=dbl_fat_nextcluster(inode->i_sb,cluster,cvfnr,NULL);
            if(cluster==-1)
            { cluster=allocate_next_cluster(inode->i_sb,lastcluster,cvfnr);
              if(cluster<=0)break;
              length=SECTOR_SIZE;
            }
            else need_cluster=1;
            offset=0;
          }
        }
        
        /* check whether there's a rest to be written */
        if(offset)
        { sektor= lastcluster==0 ? 0 :
	             dbl_mdfat_cluster2sector(inode->i_sb,lastcluster,cvfnr);

          if(dmsdos_write_cluster(inode->i_sb,clusterd,length,cluster,
	                          sektor,cvfnr,uc)<0)
	  { printk("DMSDOS: write_file: cluster write failed!\n");
	  }
	}
        
        FREE(clusterd);
        
	if(filp->f_pos>inode->i_size)
	{ inode->i_size=filp->f_pos;
	  inode->i_dirt=1;
	}
	
	return geschrieben;
}

void dmsdos_truncatex(struct inode *inode)
{
	int clusternr; 
	int size;
	int clustersize;
	int cvfnr;
	int nextcluster;
	int newval;

	cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
	if(cvfnr<0)return; /* never truncate cvf file itself */
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)return;
	
	clustersize=SECTOR_SIZE*dblsb[cvfnr].s_sectperclust;
	size=inode->i_size;
	clusternr=MSDOS_I(inode)->i_start;

        if(size!=0)
        {       
          while(size>clustersize)
          { clusternr=dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,NULL);
            if(clusternr<=0)break;
            size-=clustersize;
          }
          if(clusternr>0)
          { nextcluster=dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,NULL);
            newval=0xFFFF;
            dbl_fat_nextcluster(inode->i_sb,clusternr,cvfnr,&newval);
            if(nextcluster>0)free_chain(inode->i_sb,nextcluster,cvfnr);
          }
        }
        else /* inode->i_size==0 */
        { if(clusternr>0)free_chain(inode->i_sb,clusternr,cvfnr);
          MSDOS_I(inode)->i_start=0;
        }
        
	MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
	inode->i_dirt = 1;
}

void dmsdos_put_inodex(struct inode *inode)
{
	struct inode *depend;
	struct super_block *sb;

	if (inode->i_nlink) {
		/* if (MSDOS_I(inode)->i_busy);  ??? */
		return;
	}
	inode->i_size = 0;
	dmsdos_truncatex(inode);
	depend = MSDOS_I(inode)->i_depend;
	sb = inode->i_sb;
	clear_inode(inode);
	if (depend) {
		if (MSDOS_I(depend)->i_old != inode) {
			printk("Invalid link (0x%p): expected 0x%p, got 0x%p\n",
			    depend, inode, MSDOS_I(depend)->i_old);
#ifdef __FOR_KERNEL_1_3_8x
			fat_fs_panic(sb,"...");
#else
			fs_panic(sb,"...");
#endif
			return;
		}
		MSDOS_I(depend)->i_old = NULL;
		iput(depend);
	}
}

void dmsdos_write_inodex(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *raw_entry;
	int cvfnr;
	int dblsector;
	
	cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
	dblsector=(inode->i_ino&DMSDOS_ORIGINO_MASK)>>4;

	inode->i_dirt = 0; /* must be done before any return */
	if(cvfnr<0)return;
	if(dblsb[cvfnr].s_comp==READ_ONLY)return;
	if (inode->i_ino == MSDOS_ROOT_INO || !inode->i_nlink) return;
	if (!(bh = read_dbl_sector(sb,dblsector,cvfnr))) {
		printk("dev = 0x%04X, ino = %ld\n",inode->i_dev,inode->i_ino);
		panic("dmsdos_write_inode: unable to read i-node block");
	}
	raw_entry = &((struct msdos_dir_entry *) (bh->b_data))
	    [inode->i_ino & (MSDOS_DPB-1)];
	if (S_ISDIR(inode->i_mode)) {
		raw_entry->attr = ATTR_DIR;
		raw_entry->size = 0;
	}
	else {
		raw_entry->attr = ATTR_NONE;
		raw_entry->size = CT_LE_L(inode->i_size);
	}
	raw_entry->attr |= MSDOS_MKATTR(inode->i_mode) |
	    MSDOS_I(inode)->i_attrs;
	raw_entry->start = CT_LE_L(MSDOS_I(inode)->i_start);
#ifdef __FOR_KERNEL_1_3_8x
	fat_date_unix2dos(inode->i_mtime,&raw_entry->time,&raw_entry->date);
#else
	date_unix2dos(inode->i_mtime,&raw_entry->time,&raw_entry->date);
#endif
	raw_entry->time = CT_LE_W(raw_entry->time);
	raw_entry->date = CT_LE_W(raw_entry->date);
	bh_dirty(sb,bh);
	bh_free(sb,bh);
}

int dmsdos_notify_changex(struct inode * inode,struct iattr * attr)
{  int cvfnr;

   cvfnr=(inode->i_ino>>DMSDOS_CVFNR_BITS)-1;
   if(cvfnr<0)return -EPERM; /* never change cvf file itself */	
   if(dblsb[cvfnr].s_comp==READ_ONLY)return -EPERM;
#ifdef __FOR_KERNEL_1_3_8x
   return fat_notify_change(inode,attr);
#else
   return msdos_notify_change(inode,attr);
#endif
}

/* File creation lock. This is system-wide to avoid deadlocks in rename. */
/* (rename might deadlock before detecting cross-FS moves.) */


/* MS-DOS "device special files" */

static char *reserved_names[] = {
    "CON     ","PRN     ","NUL     ","AUX     ",
    "LPT1    ","LPT2    ","LPT3    ","LPT4    ",
    "COM1    ","COM2    ","COM3    ","COM4    ",
    NULL };

/* Characters that are undesirable in an MS-DOS file name */
  
static char bad_chars[] = "*?<>|\"";
static char bad_if_strict[] = "+=,; ";

/* Formats an MS-DOS file name. Rejects invalid names. */

int dmsdos_format_name(char conv,const char *name,int len,char *res,
  int dot_dirs)
{
	char *walk,**reserved;
	unsigned char c;
	int space;

	if (IS_FREE(name)) return -EINVAL;
	if (name[0] == '.' && (len == 1 || (len == 2 && name[1] == '.'))) {
		if (!dot_dirs) return -EEXIST;
		memset(res+1,' ',10);
		while (len--) *res++ = '.';
		return 0;
	}
	space = 1; /* disallow names starting with a dot */
	c = 0;
	for (walk = res; len && walk-res < 8; walk++) {
	    	c = *name++;
		len--;
		if (conv != 'r' && strchr(bad_chars,c)) return -EINVAL;
		if (conv == 's' && strchr(bad_if_strict,c)) return -EINVAL;
  		if (c >= 'A' && c <= 'Z' && conv == 's') return -EINVAL;
		if (c < ' ' || c == ':' || c == '\\') return -EINVAL;
		if (c == '.') break;
		space = c == ' ';
		*walk = c >= 'a' && c <= 'z' ? c-32 : c;
	}
	if (space) return -EINVAL;
	if (conv == 's' && len && c != '.') {
		c = *name++;
		len--;
		if (c != '.') return -EINVAL;
	}
	while (c != '.' && len--) c = *name++;
	if (c == '.') {
		while (walk-res < 8) *walk++ = ' ';
		while (len > 0 && walk-res < MSDOS_NAME) {
			c = *name++;
			len--;
			if (conv != 'r' && strchr(bad_chars,c)) return -EINVAL;
			if (conv == 's' && strchr(bad_if_strict,c))
				return -EINVAL;
			if (c < ' ' || c == ':' || c == '\\' || c == '.')
				return -EINVAL;
			if (c >= 'A' && c <= 'Z' && conv == 's') return -EINVAL;
			space = c == ' ';
			*walk++ = c >= 'a' && c <= 'z' ? c-32 : c;
		}
		if (space) return -EINVAL;
		if (conv == 's' && len) return -EINVAL;
	}
	while (walk-res < MSDOS_NAME) *walk++ = ' ';
	for (reserved = reserved_names; *reserved; reserved++)
		if (!strncmp(res,*reserved,8)) return -EINVAL;
	return 0;
}

static struct wait_queue *creation_wait = NULL;
static creation_lock = 0;


void lock_dmsdos_creation(void)
{
	while (creation_lock) sleep_on(&creation_wait);
	creation_lock = 1;
}


void unlock_dmsdos_creation(void)
{
	creation_lock = 0;
	wake_up(&creation_wait);
}

/* Creates a directory entry (name is already formatted). */

static int dmsdos_create_entry(struct inode *dir,char *name,int is_dir,
    struct inode **result)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *de;
	int res,ino;
	int cvfnr;
	int dblsector;
	
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)return -EPERM;

	if((res=scan_dbl_dir_4_empty(sb,MSDOS_I(dir)->i_start,cvfnr))<0)
	                                                          return res;
	ino=res;
	dblsector=(res&DMSDOS_ORIGINO_MASK)>>4;
        bh=read_dbl_sector(sb,dblsector,cvfnr);
        if(bh==0)return -EIO;
        de=(struct msdos_dir_entry*) &(bh->b_data[(res&0xf)<<5]);
	/*
	 * XXX all times should be set by caller upon successful completion.
	 */
	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_dirt = 1;
	memcpy(de->name,name,MSDOS_NAME);
	memset(de->unused, 0, sizeof(de->unused));
	de->attr = is_dir ? ATTR_DIR : ATTR_ARCH;
	de->start = 0;
#ifdef __FOR_KERNEL_1_3_8x
	fat_date_unix2dos(dir->i_mtime,&de->time,&de->date);
#else
	date_unix2dos(dir->i_mtime,&de->time,&de->date);
#endif
	de->size = 0;
	bh_dirty(sb,bh);
	if ((*result = iget(dir->i_sb,ino)) != NULL)
		dmsdos_read_inode(*result);
	bh_free(sb,bh);
	if (!*result) return -EIO;
	(*result)->i_mtime = (*result)->i_atime = (*result)->i_ctime =
	    CURRENT_TIME;
	(*result)->i_dirt = 1;
	return 0;
}

static int dmsdos_create_entry_lfn(struct inode *dir,const char *name,int len,
                                   int is_dir, struct inode **result)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head *bh;
	struct msdos_dir_entry *de;
	int res,ino;
	int cvfnr;
	int dblsector;
	
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)return -EPERM;

	if((res=create_lfn_entry(sb,MSDOS_I(dir)->i_start,name,len,cvfnr))<0)
	                                                          return res;
	ino=res;
	dblsector=(res&DMSDOS_ORIGINO_MASK)>>4;
        bh=read_dbl_sector(sb,dblsector,cvfnr);
        if(bh==0)return -EIO;
        de=(struct msdos_dir_entry*) &(bh->b_data[(res&0xf)<<5]);
	/*
	 * XXX all times should be set by caller upon successful completion.
	 */
	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_dirt = 1;
	memset(de->unused, 0, sizeof(de->unused));
	de->attr = is_dir ? ATTR_DIR : ATTR_ARCH;
	de->start = 0;
#ifdef __FOR_KERNEL_1_3_8x
	fat_date_unix2dos(dir->i_mtime,&de->time,&de->date);
#else
	date_unix2dos(dir->i_mtime,&de->time,&de->date);
#endif
	de->size = 0;
	bh_dirty(sb,bh);
	if ((*result = iget(dir->i_sb,ino)) != NULL)
		dmsdos_read_inode(*result);
	bh_free(sb,bh);
	if (!*result) return -EIO;
	(*result)->i_mtime = (*result)->i_atime = (*result)->i_ctime =
	    CURRENT_TIME;
	(*result)->i_dirt = 1;
	return 0;
}

int dmsdos_createx(struct inode *dir,const char *name,int len,int mode,
	struct inode **result)
{
	struct super_block *sb = dir->i_sb;
	char msdos_name[MSDOS_NAME];
	int res;
	int cvfnr;
	
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)
	{ iput(dir);
	  return -EPERM;
	}

	if (!dir) return -ENOENT;
	
	/* ok, now decide whether short or long */
	
	if(dblsb[cvfnr].s_support_lfn&1)
	{ lock_dmsdos_creation();
	  if(scan_dbl_dir_4_lfn(sb,MSDOS_I(dir)->i_start,name,len,cvfnr,
	                        NULL)>=0)
	  { unlock_dmsdos_creation();
	    iput(dir);
	    return -EEXIST;
	  }
	  res=dmsdos_create_entry_lfn(dir,name,len,S_ISDIR(mode),result);
	  unlock_dmsdos_creation();
	  iput(dir);
	  return res;
	}
	
	if ((res = dmsdos_format_name(MSDOS_SB(dir->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
            options.name_check,
#else
	    name_check,
#endif
	    name,len,
	    msdos_name,0)) < 0) {
		iput(dir);
		return res;
	}
	lock_dmsdos_creation();
	if (scan_dbl_dir_4_filename(sb,MSDOS_I(dir)->i_start,
	                            msdos_name,RAW,cvfnr) >= 0) {
		unlock_dmsdos_creation();
		iput(dir);
		return -EEXIST;
 	}
	res = dmsdos_create_entry(dir,msdos_name,S_ISDIR(mode),result);
	unlock_dmsdos_creation();
	iput(dir);
	return res;
}

int dmsdos_mkdirx(struct inode *dir,const char *name,int len,int mode)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode,*dot;
	char msdos_name[MSDOS_NAME];
	int res,ec,i;
	int cvfnr;
	char name_ec[257];
	
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)
	{ iput(dir);
	  return -EPERM;
	}

        if(dblsb[cvfnr].s_support_lfn&1) /* longname */
        { lock_dmsdos_creation();
	  if(scan_dbl_dir_4_lfn(sb,MSDOS_I(dir)->i_start,name,len,cvfnr,
	                        NULL)>=0)
	  { unlock_dmsdos_creation();
	    iput(dir);
	    return -EEXIST;
	  }
	  if((res=dmsdos_create_entry_lfn(dir,name,len,1,&inode))<0)
	  { unlock_dmsdos_creation();
	    iput(dir);
	    return res;
	  }
        }

        else /* shortname */
        {
	  if ((res = dmsdos_format_name(MSDOS_SB(dir->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
	    options.name_check,
#else
	    name_check,
#endif
	    name,len,
	    msdos_name,0)) < 0) {
		iput(dir);
		return res;
	  }
	  lock_dmsdos_creation();
	  if (scan_dbl_dir_4_filename(sb,MSDOS_I(dir)->i_start,
	                              msdos_name,RAW,cvfnr) >= 0) {
		unlock_dmsdos_creation();
		iput(dir);
		return -EEXIST;
 	  }
	  if ((res = dmsdos_create_entry(dir,msdos_name,1,&inode)) < 0) {
		unlock_dmsdos_creation();
		iput(dir);
		return res;
	  }
	} /* now for both long and short name */
	
	dir->i_nlink++;
	inode->i_nlink = 2; /* no need to mark them dirty */
	MSDOS_I(inode)->i_busy = 1; /* prevent lookups */
	
	if ((res = allocate_first_dir_cluster(sb,cvfnr)) < 0) goto mkdir_error;
	MSDOS_I(inode)->i_start=res;
	inode->i_size=dblsb[cvfnr].s_sectperclust*SECTOR_SIZE;
	inode->i_dirt=1;
	if ((res = dmsdos_create_entry(inode,MSDOS_DOT,1,&dot)) < 0)
		goto mkdir_error;
	dot->i_size = inode->i_size; /* doesn't grow in the 2nd create_entry */
	MSDOS_I(dot)->i_start = MSDOS_I(inode)->i_start;
	dot->i_nlink = inode->i_nlink;
	dot->i_dirt = 1;
	iput(dot);
	if ((res = dmsdos_create_entry(inode,MSDOS_DOTDOT,1,&dot)) < 0)
		goto mkdir_error;
	unlock_dmsdos_creation();
	dot->i_size = dir->i_size;
	MSDOS_I(dot)->i_start = MSDOS_I(dir)->i_start;
	dot->i_nlink = dir->i_nlink;
	dot->i_dirt = 1;
	MSDOS_I(inode)->i_busy = 0;
	iput(dot);
	iput(inode);
	iput(dir);
	return 0;
mkdir_error:
	iput(inode);
	if ((ec=dmsdos_rmdirx(dir,name,len)) < 0)
	{ for(i=0;i<len;++i)name_ec[i]=name[i];
	  name_ec[len]='\0';
	  printk("DMSDOS: rmdir %s in mkdir failed with error code %d - check filesystem\n",
	         name_ec,ec);
#ifdef __FOR_KERNEL_1_3_8x
		fat_fs_panic(dir->i_sb,"rmdir in mkdir failed");
#else
		fs_panic(dir->i_sb,"rmdir in mkdir failed");
#endif
        }
	unlock_dmsdos_creation();
	return res;
}

/* tests whether dir empty */
static int dmsdos_empty(struct inode *dir)
{
	loff_t pos;
	unsigned char buf[32];
	struct msdos_dir_entry *de;

	if (dir->i_count > 1)
		return -EBUSY;
	if (MSDOS_I(dir)->i_start) { /* may be zero in mkdir */
		pos=0;
		while (dmsdos_get_entry(dir,&pos,buf) > -1)
		{       de=(struct msdos_dir_entry*) buf;
			if (!IS_FREE(de->name) && strncmp(de->name,MSDOS_DOT,
			    MSDOS_NAME) && strncmp(de->name,MSDOS_DOTDOT,
			    MSDOS_NAME))
				return -ENOTEMPTY;		
		}
	}
	return 0;
}

int dmsdos_rmdirx(struct inode *dir,const char *name,int len)
{
	struct super_block *sb = dir->i_sb;
	int res,ino;
	struct buffer_head *bh;
	struct inode *inode;
	int cvfnr;

	bh = NULL;
	inode = NULL;
	res = -EPERM;
	if (name[0] == '.' && (len == 1 || (len == 2 && name[1] == '.')))
		goto rmdir_done;
		
	if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)
	{ iput(dir); 
	  return -EPERM;
	}
	
	/*                                
	if ((res = scan_dbl_dir_4_filename(sb,MSDOS_I(dir)->i_start,
	                               name,len,cvfnr)) < 0) goto rmdir_done;
	*/
	
	if ((res = scan_dbl_dir_4_lfn(sb,MSDOS_I(dir)->i_start,
	                       name,len,cvfnr,NULL)) < 0) goto rmdir_done;

	ino=res; /* ino of entry that is being deleted */
	res = -ENOENT;
	if (!(inode = iget(dir->i_sb,ino))) 
	{ printk("DMSDOS: rmdir: couldn't get inode\n");
	  goto rmdir_done;
	}
	res = -ENOTDIR;
	if (!S_ISDIR(inode->i_mode)) goto rmdir_done;
	res = -EBUSY;
	if (dir->i_dev != inode->i_dev || dir == inode) goto rmdir_done;
	res = dmsdos_empty(inode);
	if (res)
		goto rmdir_done;
	inode->i_nlink = 0;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_nlink--;
	inode->i_dirt = dir->i_dirt = 1;
	/*
	bh=read_dbl_sector(sb,(ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(bh==NULL){res=-EIO;goto rmdir_done;}
	bh->b_data[32*(ino&0xF)] = DELETED_FLAG;
	bh_dirty(sb,bh);
	bh_free(sb,bh);
	res = 0;
	*/
	res=remove_lfn_entry(sb,MSDOS_I(dir)->i_start,name,len,cvfnr);
rmdir_done:
	iput(dir);
	iput(inode);
	return res;
}

static int dmsdos_unlink_x(
	struct inode *dir,
	const char *name,
	int len,
	int nospc)	/* Flag special file ? */
{
	struct super_block *sb = dir->i_sb;
	int res,ino;
	struct buffer_head *bh;
	struct inode *inode;
	int cvfnr;

	bh = NULL;
	inode = NULL;

        if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[cvfnr].s_comp==READ_ONLY)
	{ iput(dir);
	  return -EPERM;
	}

        /*
	if ((res = scan_dbl_dir_4_filename(sb,MSDOS_I(dir)->i_start,
	                                   name,len,cvfnr)) < 0)
	                                	goto unlink_done;
	*/
	
	if ((res = scan_dbl_dir_4_lfn(sb,MSDOS_I(dir)->i_start,
	                              name,len,cvfnr,NULL)) < 0)
	                                        goto unlink_done;
	                                                                                                   
	ino=res;
	
	if (!(inode = iget(dir->i_sb,ino))) {
		res = -ENOENT;
		goto unlink_done;
	}
	if (!S_ISREG(inode->i_mode) && nospc){
		res = -EPERM;
		goto unlink_done;
	}
	inode->i_nlink = 0;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	MSDOS_I(inode)->i_busy = 1;
	inode->i_dirt = dir->i_dirt = 1;
	/*
	bh=read_dbl_sector(sb,(ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(bh==NULL){res=-EIO;goto unlink_done;}
	bh->b_data[32*(ino&0xf)] = DELETED_FLAG;
	bh_dirty(sb,bh);
	bh_free(sb,bh);
	res=0;
	*/
	res=remove_lfn_entry(sb,MSDOS_I(dir)->i_start,name,len,cvfnr);
unlink_done:
	iput(inode);
	iput(dir);
	return res;
}

int dmsdos_unlinkx(struct inode *dir,const char *name,int len)
{
	return dmsdos_unlink_x (dir,name,len,1);
}
/*
	Special entry for umsdos
*/
int dmsdos_unlink_umsdos(struct inode *dir,const char *name,int len)
{       if(dostest(dir)) return msdos_unlink_umsdos(dir,name,len);
	return dmsdos_unlink_x (dir,name,len,0);
}

static int rename_same_dir(struct inode *dir, const char*new_name, int new_len,
       char *new_msdos_name,int old_ino,int cvfnr)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head *new_bh,*old_bh;
	struct msdos_dir_entry *new_de,*old_de;
	struct inode *new_inode,*old_inode;
	int new_ino,exists,error;

	exists = (new_ino=scan_dbl_dir_4_filename(sb,
	                                          MSDOS_I(dir)->i_start,
	                                          new_msdos_name/*new_name*/,
	                                          RAW/*new_len*/,
	                                          cvfnr)
	         )  >= 0;
	old_bh=read_dbl_sector(sb,(old_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(old_bh==NULL)return -EIO;
	old_de=(struct msdos_dir_entry*) &(old_bh->b_data[32*(old_ino&0xf)]);
	
	if (*(unsigned char *) old_de->name == DELETED_FLAG)
	        { bh_free(sb,old_bh);
	          return -ENOENT;
	        }
	        
	if (exists) {
		if (!(new_inode = iget(dir->i_sb,new_ino))) 
		{ bh_free(sb,old_bh);
		  return -EIO;
		}

		error = S_ISDIR(new_inode->i_mode) ? (old_de->attr & ATTR_DIR) ?
		    dmsdos_empty(new_inode) : -EPERM : (old_de->attr & ATTR_DIR)
		    ? -EPERM : 0;
		    
		if (error) {
			iput(new_inode);
			bh_free(sb,old_bh);
			return error;
		}
		if (S_ISDIR(new_inode->i_mode)) {
			dir->i_nlink--;
			dir->i_dirt = 1;
		}
		
		new_inode->i_nlink = 0;
		MSDOS_I(new_inode)->i_busy = 1;
		new_inode->i_dirt = 1;
		new_bh=read_dbl_sector(sb,(new_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
		if(new_bh==NULL)
		{ iput(new_inode);
		  bh_free(sb,old_bh);
		  return -EIO;
		}
		new_de=(struct msdos_dir_entry*)
		                        &(new_bh->b_data[32*(new_ino&0xf)]);
		new_de->name[0] = DELETED_FLAG;
		bh_dirty(sb,new_bh);
		iput(new_inode);
		bh_free(sb,new_bh);
	}
	memcpy(old_de->name,new_msdos_name,MSDOS_NAME);
	bh_dirty(sb,old_bh);
	bh_free(sb,old_bh);

/* this update code seems to be always necessary in 1.3.8x kernels -- why? */
#ifndef __FOR_KERNEL_1_3_8x
	if (MSDOS_SB(dir->i_sb)->conversion == 'a') /* update binary info */
#endif
		if ((old_inode = iget(dir->i_sb,old_ino)) != NULL) {
			dmsdos_read_inode(old_inode);
			iput(old_inode);
		}
	return 0;
}

static int rename_diff_dir(struct inode *old_dir,struct inode* new_dir,
        const char *new_name,int new_len,char *new_msdos_name,int old_ino,
        int cvfnr)
{
	struct super_block *sb = old_dir->i_sb;
	struct buffer_head *new_bh,*free_bh,*dotdot_bh,*old_bh;
	struct msdos_dir_entry *new_de,*free_de,*dotdot_de,*old_de;
	struct inode *old_inode,*new_inode,*free_inode,*dotdot_inode,*walk;
	int new_ino,free_ino,dotdot_ino;
	int error,exists,ino;
	int dotdotsector;

	if (old_dir->i_dev != new_dir->i_dev) return -EINVAL;
	if (old_ino == new_dir->i_ino) return -EINVAL;
	if (!(walk = iget(new_dir->i_sb,new_dir->i_ino))) return -EIO;
	
	/* the following loop returns -EINVAL if new_dir is anywhere a
	   sub-(sub-sub-...)directory if old_ino (not old_dir)
	   ...no problem if old_ino is in fact a file, ok */
	while (walk->i_ino != MSDOS_ROOT_INO) {
		ino = dmsdos_parent_ino(walk);
		iput(walk);
		if (ino < 0) return ino;
		if (ino == old_ino) return -EINVAL;
		if (!(walk = iget(new_dir->i_sb,ino))) return -EIO;
	}
	iput(walk);
	
	/* getting free entry in new_dir */
	error=scan_dbl_dir_4_empty(sb,MSDOS_I(new_dir)->i_start,cvfnr);
	if(error<0)return error;
	free_ino=error;
	
	exists = (new_ino=scan_dbl_dir_4_filename(sb,
	                                          MSDOS_I(new_dir)->i_start,
	                                          new_msdos_name/*new_name*/,
	                                          RAW/*new_len*/,
	                                          cvfnr)
	         ) >= 0;
	         
	if (!(old_inode = iget(old_dir->i_sb,old_ino))) return -EIO;
	
	if(dbltest(old_inode))      /* cannot move dbl root dir */
	        { iput(old_inode);
	          return -EPERM;
	        }
	
	old_bh=read_dbl_sector(sb,(old_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(old_bh==NULL){iput(old_inode);return -EIO;}
	old_de=(struct msdos_dir_entry*) &(old_bh->b_data[32*(old_ino&0xf)]);
	
	if (*(unsigned char *) old_de->name == DELETED_FLAG) {
		iput(old_inode);
		bh_free(sb,old_bh);
		return -ENOENT;
	}
	
	new_inode = NULL; /* to make GCC happy */
	if (exists) {
		if (!(new_inode = iget(new_dir->i_sb,new_ino))) {
			iput(old_inode);
			bh_free(sb,old_bh);
			return -EIO;
		}
		
		error = S_ISDIR(new_inode->i_mode) ? (old_de->attr & ATTR_DIR) ?
		    dmsdos_empty(new_inode) : -EPERM : (old_de->attr & ATTR_DIR)
		    ? -EPERM : 0;
		if (error) {
			iput(new_inode);
			iput(old_inode);
			bh_free(sb,old_bh);
			return error;
		}
		
		new_inode->i_nlink = 0;
		MSDOS_I(new_inode)->i_busy = 1;
		new_inode->i_dirt = 1;
		new_bh=read_dbl_sector(sb,(new_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
		if(new_bh==NULL)
		{ iput(new_inode);
		  iput(old_inode);
		  bh_free(sb,old_bh);
		  return -EIO;
		}
		new_de=(struct msdos_dir_entry*)
		                         &(new_bh->b_data[32*(new_ino&0xf)]);
		new_de->name[0] = DELETED_FLAG;
		bh_dirty(sb,new_bh);
		bh_free(sb,new_bh);
	}
	
	free_bh=read_dbl_sector(sb,(free_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(free_bh==NULL)
	{ if(exists)iput(new_inode);
	  bh_free(sb,old_bh);
	  return -EIO;
	}
	free_de=(struct msdos_dir_entry*)&(free_bh->b_data[32*(free_ino&0xf)]);
	memcpy(free_de,old_de,sizeof(struct msdos_dir_entry));
	memcpy(free_de->name,new_msdos_name,MSDOS_NAME);
	if (!(free_inode = iget(new_dir->i_sb,free_ino))) {
		free_de->name[0] = DELETED_FLAG;
/*  Don't mark free_bh as dirty. Both states are supposed to be equivalent. */
		bh_free(sb,free_bh);
		if (exists) {
			iput(new_inode);
		}
		bh_free(sb,old_bh);
		return -EIO;
	}
	if (exists && S_ISDIR(new_inode->i_mode)) {
		new_dir->i_nlink--;
		new_dir->i_dirt = 1;
	}
	dmsdos_read_inode(free_inode);
	MSDOS_I(old_inode)->i_busy = 1;
	/*cache_inval_inode(old_inode);*/
	old_inode->i_dirt = 1;
	old_de->name[0] = DELETED_FLAG;
	bh_dirty(sb,old_bh);
	bh_dirty(sb,free_bh);
	bh_free(sb,old_bh);
	bh_free(sb,free_bh);
	if (!exists) iput(free_inode);
	else {
		MSDOS_I(new_inode)->i_depend = free_inode;
		MSDOS_I(free_inode)->i_old = new_inode;
		/* free_inode is put when putting new_inode */
		iput(new_inode);
		/*bh_free(sb,new_bh); is already done above */
	}
	
	if (S_ISDIR(old_inode->i_mode)) {
	        /* update .. entry in old_inode which is a dir inode */
	        
	        dotdotsector=dbl_mdfat_cluster2sector(sb,
	                             MSDOS_I(old_inode)->i_start,cvfnr);
	        dotdot_bh=read_dbl_sector(sb,dotdotsector,cvfnr);
	        if(dotdot_bh==NULL){error=-EIO;goto rename_done;}
	        dotdot_de=(struct msdos_dir_entry*)
	                                         &(dotdot_bh->b_data[32]);
	        
		dotdot_ino=(dotdotsector<<4)+1+((cvfnr+1)<<DMSDOS_CVFNR_BITS);
		if (!(dotdot_inode = iget(old_inode->i_sb,dotdot_ino))) {
			error = -EIO;
			goto rename_done;
		}
		dotdot_de->start = MSDOS_I(dotdot_inode)->i_start =
		    MSDOS_I(new_dir)->i_start;
		dotdot_inode->i_dirt = 1;
		bh_dirty(sb,dotdot_bh);
		old_dir->i_nlink--;
		new_dir->i_nlink++;
		/* no need to mark them dirty */
		dotdot_inode->i_nlink = new_dir->i_nlink;
		iput(dotdot_inode);
		bh_free(sb,dotdot_bh);
	}
	error = 0;
rename_done:
	iput(old_inode);
	return error;
}

/* the same for long filenames */
/* no difference for same directories ... */

static int rename_dir_lfn(struct inode *old_dir,struct inode* new_dir,
                            int cvfnr,int old_ino,
                            const char *old_name, int old_len,
                            const char *new_name, int new_len)
{
	struct super_block *sb = old_dir->i_sb;
	struct buffer_head /* *new_bh, */ *free_bh,*dotdot_bh,*old_bh;
	struct msdos_dir_entry /* *new_de, */ *free_de,*dotdot_de,*old_de;
	struct inode *old_inode,*new_inode,*free_inode,*dotdot_inode,*walk;
	int new_ino,free_ino,dotdot_ino;
	int error,exists,ino;
	int dotdotsector;
	unsigned char new_msdos_name[MSDOS_NAME];

	if (old_dir->i_dev != new_dir->i_dev) return -EINVAL;
	if (old_ino == new_dir->i_ino) return -EINVAL;
	
	if (old_dir->i_ino != new_dir->i_ino)
	{ /* this is the only difference between same_dir and diff_dir */
	  if (!(walk = iget(new_dir->i_sb,new_dir->i_ino))) return -EIO;
	
  	  /* the following loop returns -EINVAL if new_dir is anywhere a
	     sub-(sub-sub-...)directory if old_ino (not old_dir)
	     ...no problem if old_ino is in fact a file, ok */
	  while (walk->i_ino != MSDOS_ROOT_INO) {
		ino = dmsdos_parent_ino(walk);
		iput(walk);
		if (ino < 0) return ino;
		if (ino == old_ino) return -EINVAL;
		if (!(walk = iget(new_dir->i_sb,ino))) return -EIO;
	  }
	  iput(walk);
	}
	
	/* getting free entry in new_dir and write lfn ATTR_VOLs */
	error=create_lfn_entry(sb,MSDOS_I(new_dir)->i_start,
	                       new_name,new_len,cvfnr);
	if(error<0)return error;
	free_ino=error;
	
	/* this _doesn't_ find the above created entry 
	   because create_lfn_entry marks even the short name entry
	   with ATTR_VOL */
	exists = (new_ino=scan_dbl_dir_4_lfn(sb,
	                                          MSDOS_I(new_dir)->i_start,
	                                          new_name,
	                                          new_len,
	                                          cvfnr,NULL)
	         ) >= 0;
	         
	if (!(old_inode = iget(old_dir->i_sb,old_ino))) return -EIO;
	
	if(dbltest(old_inode))      /* cannot move dbl root dir */
	        { iput(old_inode);
	          return -EPERM;
	        }
	
	old_bh=read_dbl_sector(sb,(old_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(old_bh==NULL){iput(old_inode);return -EIO;}
	old_de=(struct msdos_dir_entry*) &(old_bh->b_data[32*(old_ino&0xf)]);
	
	if (*(unsigned char *) old_de->name == DELETED_FLAG) {
		iput(old_inode);
		bh_free(sb,old_bh);
		return -ENOENT;
	}
	
	new_inode = NULL; /* to make GCC happy */
	if (exists) {
		if (!(new_inode = iget(new_dir->i_sb,new_ino))) {
			iput(old_inode);
			bh_free(sb,old_bh);
			return -EIO;
		}
		
		error = S_ISDIR(new_inode->i_mode) ? (old_de->attr & ATTR_DIR) ?
		    dmsdos_empty(new_inode) : -EPERM : (old_de->attr & ATTR_DIR)
		    ? -EPERM : 0;
		if (error) {
			iput(new_inode);
			iput(old_inode);
			bh_free(sb,old_bh);
			return error;
		}
		
		new_inode->i_nlink = 0;
		MSDOS_I(new_inode)->i_busy = 1;
		new_inode->i_dirt = 1;
		/*
		new_bh=read_dbl_sector(sb,(new_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
		if(new_bh==NULL)
		{ iput(new_inode);
		  iput(old_inode);
		  bh_free(sb,old_bh);
		  return -EIO;
		}
		new_de=(struct msdos_dir_entry*)
		                         &(new_bh->b_data[32*(new_ino&0xf)]);
		new_de->name[0] = DELETED_FLAG;
		bh_dirty(sb,new_bh);
		bh_free(sb,new_bh);
		*/
		if(remove_lfn_entry(sb,MSDOS_I(new_inode)->i_start,
		                 new_name,new_len,cvfnr)<0)
		{ iput(new_inode);
		  iput(old_inode);
		  bh_free(sb,old_bh);
		  return -EIO;
		}
	}
	
	free_bh=read_dbl_sector(sb,(free_ino&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
	if(free_bh==NULL)
	{ if(exists)iput(new_inode);
	  bh_free(sb,old_bh);
	  return -EIO;
	}
	free_de=(struct msdos_dir_entry*)&(free_bh->b_data[32*(free_ino&0xf)]);
	
	/* copy everything except (short) name */
	/* name retten */
	memcpy(new_msdos_name,free_de->name,MSDOS_NAME);
	memcpy(free_de,old_de,sizeof(struct msdos_dir_entry));
	memcpy(free_de->name,new_msdos_name,MSDOS_NAME);
	        
	if (!(free_inode = iget(new_dir->i_sb,free_ino))) {
		free_de->name[0] = DELETED_FLAG;
/*  Don't mark free_bh as dirty. Both states are supposed to be equivalent. */
		bh_free(sb,free_bh);
		if (exists) {
			iput(new_inode);
		}
		bh_free(sb,old_bh);
		return -EIO;
	}
	if (exists && S_ISDIR(new_inode->i_mode)) {
		new_dir->i_nlink--;
		new_dir->i_dirt = 1;
	}
	dmsdos_read_inode(free_inode);
	MSDOS_I(old_inode)->i_busy = 1;
	/*cache_inval_inode(old_inode);*/
	bh_free(sb,old_bh);
	remove_lfn_entry(sb,MSDOS_I(old_dir)->i_start,
	                 old_name,old_len,cvfnr);
	old_inode->i_dirt = 1;
	/*
	old_de->name[0] = DELETED_FLAG;
	bh_dirty(sb,old_bh);
	*/
	bh_dirty(sb,free_bh);
	bh_free(sb,free_bh);
	if (!exists) iput(free_inode);
	else {
		MSDOS_I(new_inode)->i_depend = free_inode;
		MSDOS_I(free_inode)->i_old = new_inode;
		/* free_inode is put when putting new_inode */
		iput(new_inode);
		/*bh_free(sb,new_bh); is already done above */
	}
	
	if (S_ISDIR(old_inode->i_mode)) {
	        /* update .. entry in old_inode which is a dir inode */
	        
	        dotdotsector=dbl_mdfat_cluster2sector(sb,
	                             MSDOS_I(old_inode)->i_start,cvfnr);
	        dotdot_bh=read_dbl_sector(sb,dotdotsector,cvfnr);
	        if(dotdot_bh==NULL){error=-EIO;goto rename_done;}
	        dotdot_de=(struct msdos_dir_entry*)
	                                         &(dotdot_bh->b_data[32]);
	        
		dotdot_ino=(dotdotsector<<4)+1+((cvfnr+1)<<DMSDOS_CVFNR_BITS);
		if (!(dotdot_inode = iget(old_inode->i_sb,dotdot_ino))) {
			error = -EIO;
			goto rename_done;
		}
		dotdot_de->start = MSDOS_I(dotdot_inode)->i_start =
		    MSDOS_I(new_dir)->i_start;
		dotdot_inode->i_dirt = 1;
		bh_dirty(sb,dotdot_bh);
		old_dir->i_nlink--;
		new_dir->i_nlink++;
		/* no need to mark them dirty */
		dotdot_inode->i_nlink = new_dir->i_nlink;
		iput(dotdot_inode);
		bh_free(sb,dotdot_bh);
	}
	error = 0;
rename_done:
	iput(old_inode);
	return error;
}

int dmsdos_renamex(struct inode *old_dir,const char *old_name,int old_len,
	struct inode *new_dir,const char *new_name,int new_len
#ifdef __FOR_KERNEL_2_0_1
	, int must_be_dir
#endif
	)
{
	struct super_block *sb = old_dir->i_sb;
	char old_msdos_name[MSDOS_NAME],new_msdos_name[MSDOS_NAME];
	int old_ino,error;
	int old_cvfnr;
	int new_cvfnr;
	
	/* cannot move from msdos to dmsdos and vice versa: 
	   different partitions */
	/* maybe I have to implement a copy and delete here */
	if(dostest(old_dir)||dostest(new_dir))
	{ iput(old_dir);
	  iput(new_dir);
	  return -EINVAL;
	}
	
	if((old_cvfnr=dbltest(old_dir))!=0)old_cvfnr=(old_cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else old_cvfnr=(old_dir->i_ino>>DMSDOS_CVFNR_BITS)-1;

        if(dblsb[old_cvfnr].s_comp==READ_ONLY)
        { iput(old_dir);
          iput(new_dir);
          return -EPERM;
        }
	
	if((new_cvfnr=dbltest(new_dir))!=0)new_cvfnr=(new_cvfnr>>DMSDOS_CVFNR_BITS)-1;
	else new_cvfnr=(new_dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
	
	if(dblsb[new_cvfnr].s_comp==READ_ONLY)
	{ iput(old_dir);
	  iput(new_dir);
	  return -EPERM;
	}
	
	/* cannot move across cvfs: different partitions */
	/* maybe a *compressed* copy and delete ??? */
	if(old_cvfnr!=new_cvfnr)
	{ iput(old_dir);
	  iput(new_dir);
	  return -EINVAL;
	}
	
	/* now check for lfn */
	
	if(dblsb[old_cvfnr].s_support_lfn&1)
	{ if((error=scan_dbl_dir_4_lfn(sb,MSDOS_I(old_dir)->i_start,
	                  old_name,old_len,old_cvfnr,NULL))<0)goto rename_done;
	  old_ino=error;
	  lock_dmsdos_creation();
	  if(old_dir==new_dir)
	    error=(old_len==new_len&&strncmp(old_name,new_name,old_len)==0) ?
	                    0 :
	                    rename_dir_lfn(old_dir,new_dir,old_cvfnr,old_ino,
	                           old_name,old_len,new_name,new_len);
	  else
	    error=rename_dir_lfn(old_dir,new_dir,old_cvfnr,old_ino,
	                           old_name,old_len,new_name,new_len);
	  unlock_dmsdos_creation();
	  goto rename_done; 
	}
	
	if ((error = dmsdos_format_name(MSDOS_SB(old_dir->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
            options.name_check,
#else
	    name_check,
#endif
	    old_name,old_len,old_msdos_name,1)) < 0) goto rename_done;
	if ((error = dmsdos_format_name(MSDOS_SB(new_dir->i_sb)->
#ifdef __FOR_KERNEL_1_3_8x
            options.name_check,
#else
	    name_check,
#endif
	    new_name,new_len,new_msdos_name,0)) < 0) goto rename_done;
	if ((error = scan_dbl_dir_4_filename(sb,MSDOS_I(old_dir)->i_start,
	                  old_msdos_name/*old_name*/,RAW/*old_len*/,old_cvfnr)) < 0) goto rename_done;
	old_ino=error;
	lock_dmsdos_creation();
	if (old_dir == new_dir)
		error = strncmp(old_msdos_name,new_msdos_name,MSDOS_NAME)==0 ?
		           0 :
		           rename_same_dir(old_dir,new_name,new_len,
		                           new_msdos_name,old_ino,old_cvfnr);
	else error = rename_diff_dir(old_dir,new_dir,new_name,new_len,
		                     new_msdos_name,old_ino,old_cvfnr);
	unlock_dmsdos_creation();
rename_done:
	iput(old_dir);
	iput(new_dir);
	return error;
}
